#include <stdio.h>
#include <shmem.h>
#include <math.h>
#include "pgmio.h"

#define nx 360
#define ny 240
#define numpes 8
#define nxlocal (nx/numpes)
#define nylocal ny
#define maxloop 100000
#define nr 1

#define MAX(X,Y) ((X) > (Y) ? (X) : (Y))

double bigpic[nx][ny];
double oldpic[nxlocal+2][nylocal+2];
double newpic[nxlocal+2][nylocal+2];
double edge[nxlocal+2][nylocal+2];

double deltasq, deltasqsum;

long psync[_SHMEM_BCAST_SYNC_SIZE];

int main(void)
{
  
  int ix, iy, iloop;
  int pe, mype, rtpe, ltpe;
  double globalsum;
  double delta;
  double deltamin = 0.1;

  double *pwrk;
  int pwrkSize;

  shmem_init();

  mype = shmem_my_pe();

  if (numpes != shmem_n_pes())
  {

    if (mype == 0)
    {
      printf("Error: compiled for %d PE(s), but running on %d\n", numpes, shmem_n_pes());
    }

    shmem_finalize();

    return -1;

  }

  if (mype == 0)
  {
    printf("Running with %d PE(s)\n\n",numpes);
    printf("Performing %d iterations\n\n",maxloop);
  }

  printf("Hello from PE %d out of %d\n",mype,numpes);

  /* Allocate and initialise work arrays needed for later reduction */
  pwrkSize = MAX((nx*ny)/2 + 1, _SHMEM_REDUCE_MIN_WRKDATA_SIZE);
  pwrk = (double *) shmalloc(pwrkSize*sizeof(double));

  for (ix=0; ix < _SHMEM_BCAST_SYNC_SIZE; ix++)
  {
    psync[ix] = _SHMEM_SYNC_VALUE;
  }

  for (ix=0; ix < (nxlocal+2); ix++)
  {
    for (iy=0; iy < (nylocal+2); iy++)
    {
      oldpic[ix][iy] = 0.0;
      newpic[ix][iy] = 0.0;
      edge[ix][iy] = 0.0;
    }
  }

  /* Need to synchronise to ensure smallpic is initialised on all PEs
     before the master PE reads them. */

  shmem_barrier_all();

  /* Read the picture on the master and copy to every PE's edge */

  if (mype == 0)
  {

    pgmread("edge360x240.pgm", bigpic, nx, ny);

    /* Mapping done in strips with multiple put calls as edge has a halo.
     A single block put would overlap that halo. */

    for (pe=0; pe < numpes; pe++)
    {
      for(ix = pe*nxlocal; ix < (pe+1)*nxlocal; ix++)
      {
        shmem_double_put(&edge[ (ix % nxlocal)+1 ][1], &bigpic[ix][0], nylocal, pe);
      }
    }

  }

  /* Overwrite bigpic to ensure that we have to do the output correctly */
  for (ix=0; ix < nx; ix++)
  {
    for (iy=0; iy < ny; iy++)
    {
      bigpic[ix][iy] = -1.0;
    }
  }

  /* Need to synchronise to ensure the master has completed initialising
     edge on all PEs before any of them try to read the halos. */

  shmem_barrier_all();

  for (ix=0; ix < (nxlocal+2); ix++)
  {
    for (iy=0; iy < (nylocal+2); iy++)
    {
      oldpic[ix][iy] = edge[ix][iy];
    }
  }

  /* Compute indices of neighbouring PEs */
  rtpe = mype + 1;
  ltpe = mype - 1;

  /* Need to synchronise to ensure oldpic has been initialised on all PEs
  before any PEs try to access its halos on their neighbours */

  shmem_barrier_all();

  for (iloop=1; iloop <= maxloop; iloop++)
  {
    /* Do halo swap, taking care at the picture boundaries */
  
    if (rtpe < numpes)
    {
      shmem_double_get(&oldpic[nxlocal+1][1], &oldpic[1][1], nylocal, rtpe);
    }

    if (ltpe >= 0)
    {
      shmem_double_get(&oldpic[0][1], &oldpic[nxlocal][1], nylocal, ltpe);
    }

    /* Now update the local values of newpic */
  
    for (ix=1; ix <= nxlocal; ix++)
    {
      for (iy=1; iy <= nylocal; iy++)
      {
        newpic[ix][iy] = 0.25*(   oldpic[ix-1][iy] + oldpic[ix+1][iy]
                                + oldpic[ix][iy-1] + oldpic[ix][iy+1]
                                - edge[ix][iy]                      );
      }
    }

    /* Now compute the delta parameter locally */
    deltasq = 0.0;
    for (ix=1; ix <= nxlocal; ix++)
    {
      for (iy=1; iy <= nylocal; iy++)
      {
        deltasq += pow(newpic[ix][iy] - oldpic[ix][iy], 2);
      }
    }

    deltasq = deltasq / (double)(nx*ny);

    /* Sum up the values across PEs and take the square root */
    shmem_double_sum_to_all(&deltasqsum, &deltasq, nr, 0, 0, numpes,
                            pwrk, psync);

    delta = sqrt(deltasqsum);

    if (mype == 0)
    {
      printf("Iteration %d delta = %G\n",iloop,delta);
    }

    /* Exit if convergence criterion is reached */

    if (delta < deltamin)
    {
      break;
    }

    /* Need to synchronise to ensure that all PEs have finished reading the
       oldpic halo values on this PE before overwriting it with newpic */

    shmem_barrier_all();

    /* Prepare for next iteration by setting oldpic = newpic */

    for (ix=1; ix <= nxlocal; ix++)
    {
      for (iy=1; iy <= nylocal; iy++)
      {
        oldpic[ix][iy] = newpic[ix][iy];
      }
    }

    /* Need to synchronise to ensure that all PEs have finished updating
       their oldpic arrays before this PE reads any halo data from them at
       the start of the next iteration */

    shmem_barrier_all();

  }

  /* Now write the output on the master PE.
     No synchronisation required as due to sync all at the end of each loop */

  if (mype == 0)
  {

    /* Copy newpic back to bigpic */
    for (pe=0; pe < numpes; pe++)
    {
      for(ix = pe*nxlocal; ix < (pe+1)*nxlocal; ix++)
      {
        shmem_double_get(&bigpic[ix][0], &newpic[ (ix % nxlocal)+1 ][1], nylocal, pe);
      }
    }

    pgmwrite("output.pgm", bigpic, nx, ny);

  }

  /* Clean up */

  shfree(pwrk);

  shmem_finalize();

  return 0;
}

